import numpy as np


'''
This script contains Steve's MultipoleExpansion code for calculating the power spectrum.
'''


def getpk(datgrid,w,W,nx,ny,nz,lx,ly,lz,kbins):
    '''
    Obtain the 3D angle-averaged power spectrum for auto-correlation.
    
    ------
    Inputs
    ------
    
    datgrid: data cube you wish to measure the power spectrum of, shape [Nx, Ny, Nz]
    w: weighting function (e.g. inverse noise weighting), shape [Nx, Ny, Nz]
    W: pixel masking function (1 for pixels to consider, 0 for pixels to ignore, shape [Nx, Ny, Nz]
    nx, ny, nz: shape of data cube [Nx, Ny, Nz]
    lx, ly, lz: data cube box dimensions, in Mpc/h
    kbins: edges of k_bins you wish to average your power spectrum into
    
    -------
    Outputs
    -------
    
    pk: 3D angle-averaged power spectrum
    nmodes: number of modes in each k bin
    
    '''
    vol = lx*ly*lz
    nc = nx*ny*nz
    pkspec = getpkspec(datgrid,nc,vol,w,W)
    pk, nmodes = binpk(pkspec,nx,ny,nz,lx,ly,lz,kbins,FullPk=False,doindep=True)
    return pk, nmodes


def getpkspec(datgrid,nc,vol,w,W):
    '''
    Estimate the 3D power spectrum of a density field (auti-correlation).
    '''
    Vcell = vol/nc
    Wmean = np.mean(W)
    W = W/np.sum(W)
    w = w/np.sum(w)
    fgrid = np.fft.rfftn(w * datgrid)
    pkspec = np.real( fgrid * np.conj(fgrid) )
    return vol * pkspec / ( nc*np.sum(W**2 * w**2) * Wmean**2 * nc**2 )


def binpk(pkspec,nx,ny,nz,lx,ly,lz,kbins,FullPk=False,doindep=True):
    '''
    Bin 3D power spectrum in angle-averaged bins.
    '''
    #print('\nBinning 3D power spectrum into angle-averaged bins...')
    kspec,muspec,indep = getkspec(nx,ny,nz,lx,ly,lz,FullPk)
    if doindep==True:
        pkspec = pkspec[indep==True]
        kspec = kspec[indep==True]
    ikbin = np.digitize(kspec,kbins)
    nkbin = len(kbins)-1
    nmodes,pk = np.zeros(nkbin,dtype=int),np.zeros(nkbin)
    for ik in range(nkbin):
        nmodes[ik] = int(np.sum(np.array([ikbin==ik+1])))
        if (nmodes[ik] > 0): #if nmodes==0 for this k then remains Pk=0
            pk[ik] = np.mean(pkspec[ikbin==ik+1])
    return pk,nmodes


def get_biascorr(fg_cov, Input, n_samples, lx, ly, lz, kbins_par, kbins, w, W):
    '''
    Obtain random samples given our foreground covariance
    
    ------
    Inputs
    ------
    
    fg_cov: your foreground covariance (shape [Nfreq, Nfreq]), the error on your foreground fit
    Input: your input data, that you are trying to foreground clean, of shape [Nx, Ny, Nz]
    n_samples: number of samples you want to draw, each sample will have shape [Nx, Ny, Nz]
    lx, ly, lz: dimensions of your data cube, in Mpc/h
    w: weighting function
    W: window function
    
    -------
    Outputs
    -------
    
    pk_corr: bias correction for 3D P(k)
    pk_corr_para: bias correction for 1D P(k_parallel)
    samples: an array of samples derived from fg_cov, total of n_samples samples, each of
        shape [Nx, Ny, Nz]
    '''
    nx, ny, nz = Input.shape
    mu = np.zeros(nz) # vector of the means, zero

    # Generate n separate samples paths from a Gaussian with mean mu and covariance fg_cov:
    samples = []
    for i in range(n_samples):
        s = np.random.multivariate_normal(mu, fg_cov, (nx, ny))# * Input.std() * noise_smoothed.std()
        samples.append(s)
    samples = np.array(samples)
    
    # Obtain 3D P(k) and 1D P(k_para) for each of these samples
    pk_corr = []
    pk_corr_para = []
    for s in samples:
        pk_corr_para.append(ParaPk(s, nx, ny, nz, lx, ly, lz, kbins_par, w, W)[0])
        pk_corr.append(getpk_noPerp(s,w,W,nx,ny,nz,lx,ly,lz,kbins)[0])
    
    # Average power spectral over all samples:
    pk_corr = np.mean(pk_corr, axis=0)
    pk_corr_para = np.mean(pk_corr_para, axis=0)
    
    return pk_corr, pk_corr_para, samples


def PerpParaPk(datgrid,nx,ny,nz,lx,ly,lz,kperpbins,kparabins,w,W):
    '''
    Return 2D image of P(k_perp,k_para)
    '''
    # Obtain two 3D arrays specifying kperp and kpara values at every point in
    #    pkspec array
    print('\nCalculating P(k_perp,k_para)...')
    pkspec = getpkspec(datgrid,nx*ny*nz,lx*ly*lz,w,W)
    kx = 2*np.pi*np.fft.fftfreq(nx,d=lx/nx)
    ky = 2*np.pi*np.fft.fftfreq(ny,d=ly/ny)
    kperp = np.sqrt(kx[:,np.newaxis]**2 + ky[np.newaxis,:]**2)
    kpara = 2*np.pi*np.fft.fftfreq(nz,d=lz/nz)[:int(nz/2)+1]
    kperp_arr = np.reshape( np.repeat(kperp,int(nz/2)+1) , (nx,ny,int(nz/2)+1) )
    kpara_arr = np.tile(kpara,(nx,ny,1))
    # Identify and remove non-independent modes
    null1,null2,indep = getkspec(nx,ny,nz,lx,ly,lz)
    pkspec = pkspec[indep==True]
    kperp_arr = kperp_arr[indep==True]
    kpara_arr = kpara_arr[indep==True]
    # Get indices where kperp and kpara values fall in bins
    ikbin_perp = np.digitize(kperp_arr,kperpbins)
    ikbin_para = np.digitize(kpara_arr,kparabins)
    nkperpbin,nkparabin = len(kperpbins)-1,len(kparabins)-1
    nmodes,pk2d = np.zeros((nkparabin,nkperpbin),dtype=int),np.zeros((nkparabin,nkperpbin))
    for i in range(nkperpbin):
        for j in range(nkparabin):
            nmodes[j,i] = int(np.sum(np.array([(ikbin_perp==i+1) & (ikbin_para==j+1)])))
            if (nmodes[j,i] > 0):
                # Average power spectrum into (kperp,kpara) cells
                pk2d[j,i] = np.mean(pkspec[(ikbin_perp==i+1) & (ikbin_para==j+1)])
    return pk2d,nmodes


def ParaPk(datgrid, nx, ny, nz, lx, ly, lz, kparabins, w, W):
    '''
    Return 1D P(k_para)
    '''
    # calculating power spectrum of overdensity grid:
    pkspec = getpkspec(datgrid,nx*ny*nz,lx*ly*lz,w,W)
    # get 1D array of available k_para modes
    kpara = 2*np.pi*np.fft.fftfreq(nz,d=lz/nz)[:int(nz/2)+1]
    # fill a cube the size of our power spectrum with copies of kpara
    kpara_arr = np.tile(kpara,(nx,ny,1))
    # identify and remove non-independent modes
    null1,null2,indep = getkspec(nx,ny,nz,lx,ly,lz)
    pkspec = pkspec[indep==True]
    kpara_arr = kpara_arr[indep==True]
    # get indices where kpara values fall in desired bins
    ikbin_para = np.digitize(kpara_arr,kparabins)
    nkparabin = len(kparabins)-1
    nmodes1d,pk1d = np.zeros(nkparabin,dtype=int),np.zeros(nkparabin)
    for i in range(nkparabin):
        nmodes1d[i] = int(np.sum(np.array([ikbin_para==i+1])))
        if (nmodes1d[i] > 0):
            # average power spectrum into kpara cells
            pk1d[i] = np.mean(pkspec[ikbin_para==i+1])
    return pk1d, nmodes1d


def PerpPk(datgrid,nx,ny,nz,lx,ly,lz,kperpbins,w,W):
    '''
    Return 1D P(k_perp)
    '''
    pkspec = getpkspec(datgrid,nx*ny*nz,lx*ly*lz,w,W)
    kx = 2*np.pi*np.fft.fftfreq(nx,d=lx/nx)
    ky = 2*np.pi*np.fft.fftfreq(ny,d=ly/ny)
    kperp = np.sqrt(kx[:,np.newaxis]**2 + ky[np.newaxis,:]**2)
    kperp_arr = np.reshape( np.repeat(kperp,int(nz/2)+1) , (nx,ny,int(nz/2)+1) )
    # Identify and remove non-independent modes
    null1,null2,indep = getkspec(nx,ny,nz,lx,ly,lz)
    pkspec = pkspec[indep==True]
    kperp_arr = kperp_arr[indep==True]
    # Get indices where kperp and kpara values fall in bins
    ikbin_perp = np.digitize(kperp_arr,kperpbins)
    nkperpbin = len(kperpbins)-1
    nmodes1d,pk1d = np.zeros((nkperpbin),dtype=int),np.zeros((nkperpbin))
    for i in range(nkperpbin):
        nmodes1d[i] = int(np.sum(np.array([ikbin_perp==i+1])))
        if (nmodes1d[i] > 0):
            # Average power spectrum into kperp cells
            pk1d[i] = np.mean(pkspec[ikbin_perp==i+1])
    return pk1d, nmodes1d


def getpk_noPerp(datgrid,w,W,nx,ny,nz,lx,ly,lz,kbins):
    '''
    Return 1D P(k_perp)
    '''
    pkspec = getpkspec(datgrid,nx*ny*nz,lx*ly*lz,w,W)
    # get 1D array of available k_para modes
    kpara = 2*np.pi*np.fft.fftfreq(nz,d=lz/nz)[:int(nz/2)+1]
    # fill a cube the size of our power spectrum with copies of kpara
    kpara_arr = np.tile(kpara,(nx,ny,1))
    # identify and remove non-independent modes
    kspec,muspec,indep = getkspec(nx,ny,nz,lx,ly,lz)
    pkspec = pkspec[indep==True]
    kpara_arr = kpara_arr[indep==True]
    kspec = kspec[indep==True]
    # get indices where kpara values fall in desired 1D bins
    nkbin = len(kbins)-1
    ikbin_para = np.digitize(kpara_arr,kbins)
    nmodes1d,pk1d = np.zeros(nkbin,dtype=int),np.zeros(nkbin)
    # same but for 3D bins:
    ikbin = np.digitize(kspec,kbins)
    nmodes3d,pk3d = np.zeros(nkbin,dtype=int),np.zeros(nkbin)
    for i in range(nkbin):
        nmodes1d[i] = int(np.sum(np.array([ikbin_para==i+1])))
        nmodes3d[i] = int(np.sum(np.array([ikbin==i+1])))
        if (nmodes1d[i] > 0):
            # average power spectrum into kpara cells
            pk1d[i] = np.mean(pkspec[ikbin_para==i+1])
    pk3d = (pk1d*nmodes1d)/nmodes3d
    return pk3d, nmodes3d


def getpk_2Dto3D(pk2d, nmodes2d, kpara_array, kperp_array, kbins):
    k2D = np.zeros((len(kpara_array), len(kperp_array)))
    pk3d = np.zeros(len(kbins)-1)
    nmodes = np.zeros(len(kbins)-1)
    for i in range(len(kperp_array)):
        for j in range(len(kpara_array)):
            k2D[j,i] = np.sqrt(kpara_array[j]**2 + kperp_array[i]**2)
    for i in range(len(kbins)-1):
        kmin, kmax = kbins[i], kbins[i+1]
        pk_i = []
        nmodes_i = []
        for k in range(len(kperp_array)):
            for j in range(len(kpara_array)):
                if (k2D[j,k] > kmin) & (k2D[j,k] < kmax):
                    if nmodes2d[j,k] > 0:
                        pk_i.append(pk2d[j,k]*nmodes2d[j,k])
                        nmodes_i.append(nmodes2d[j,k])
        pk_i = np.sum(pk_i)
        nmodes_i = np.sum(nmodes_i)
        pk3d[i] = pk_i/nmodes_i
        nmodes[i] = nmodes_i
    return pk3d, nmodes


def getkspec(nx,ny,nz,lx,ly,lz,FullPk=False):
    kx = 2.*np.pi*np.fft.fftfreq(nx,d=lx/nx)
    ky = 2.*np.pi*np.fft.fftfreq(ny,d=ly/ny)
    if FullPk==True: kz = 2.*np.pi*np.fft.fftfreq(nz,d=lz/nz)
    else: kz = 2.*np.pi*np.fft.fftfreq(nz,d=lz/nz)[:int(nz/2)+1]
    indep = getindep(nx,ny,nz)
    indep[0,0,0] = False
    if FullPk==True:
        indep = fthalftofull(nx,ny,nz,indep)
    kspec = np.sqrt(kx[:,np.newaxis,np.newaxis]**2 + ky[np.newaxis,:,np.newaxis]**2 + kz[np.newaxis,np.newaxis,:]**2)
    kspec[0,0,0] = 1.
    muspec = np.absolute(kz[np.newaxis,np.newaxis,:])/kspec
    kspec[0,0,0] = 0.
    return kspec,muspec,indep


def fthalftofull(nx,ny,nz,halfspec):
    fullspec = np.empty((nx,ny,nz))
    ixneg,iyneg,izneg = nx-np.arange(nx),ny-np.arange(ny),nz-np.arange(int(nz/2)+1,nz)
    ixneg[0],iyneg[0] = 0,0
    fullspec[:,:,:int(nz/2)+1] = halfspec
    fullspec[:,:,int(nz/2)+1:nz] = fullspec[:,:,izneg][:,iyneg][ixneg]
    return fullspec


def getindep(nx,ny,nz):
    indep = np.full((nx,ny,int(nz/2)+1),False,dtype=bool)
    indep[:,:,1:int(nz/2)] = True
    indep[1:int(nx/2),:,0] = True
    indep[1:int(nx/2),:,int(nz/2)] = True
    indep[0,1:int(ny/2),0] = True
    indep[0,1:int(ny/2),int(nz/2)] = True
    indep[int(nx/2),1:int(ny/2),0] = True
    indep[int(nx/2),1:int(ny/2),int(nz/2)] = True
    indep[int(nx/2),0,0] = True
    indep[0,int(ny/2),0] = True
    indep[int(nx/2),int(ny/2),0] = True
    indep[0,0,int(nz/2)] = True
    indep[int(nx/2),0,int(nz/2)] = True
    indep[0,int(ny/2),int(nz/2)] = True
    indep[int(nx/2),int(ny/2),int(nz/2)] = True
    return indep





import numpy as np
import re


'''
This script contains code for getting samples from a particular kernel, and the corresponding equation. 
Taken from: https://gist.github.com/rikrd/05f5ca0c31adb9203823
'''


def get_kernel_samples(kernel_name, X, n_samples, l=False, var=1.0):
    """
    Generate the multivariate random distribution of data with zero mean and covariance kernel
    specified by kernel_name, and obtain n samples.
    
    INPUTS:
    kernel_name: name of covariance kernel as in dir(GPy.kern)
    X: x data points to calculate kernel over
    n_samples: number of samples to obtain from multivariate random distribution
    l: lengthscale, always False if kernel doesn't have a lengthscale hyperparameter
    var: variance
    """
    import GPy
    # Get the kernel from GPy:
    Kernel = getattr(GPy.kern, kernel_name)
    if l==False: k = Kernel(input_dim=1)
    else: k = Kernel(input_dim=1, lengthscale=l, variance=var)
    
    X = X[:,None] # reshape X to make it n*p, they try to use 'design matrices' in GPy 
    mu = np.zeros((len(X)))# vector of the means, zero
    
    # Compute the covariance matrix associated with inputs X:
    C = k.K(X,X) 
    
    # Generate n separate samples paths from a Gaussian with mean mu and covariance C:
    Z = np.random.multivariate_normal(mu,C,n_samples)
    return Z, C


def get_equation(kern):
    """
    Retrieves the LaTeX format functional form of covariance kernel (kern).
    """
    match = re.search(r'(math::)(\r\n|\r|\n)*(?P<equation>.*)(\r\n|\r|\n)*', kern.__doc__)
    return '' if match is None else match.group('equation').strip()



from scipy.ndimage import gaussian_filter
import numpy as np


'''
This script contains code for modelling observational effects in HI IM simulations, specifically the telescope beam smoothing effect (based on Steve Cunnington's beam smoothing code). 
It also contains code for how to convert a data cube in [Nx, Ny, Nz] format to [Npix, Nz] format.

'''


def ConvolveCube(dT, zeff, lx, ly, sigma_beam, cosmo):
    '''
    Function to smooth entire data cube one slice at a time
    
    INPUTS:
    dT: data to be smoothed, in format [Nx, Ny, Nz] where Nz is frequency direction
    zeff: effective redshift of data cube
    lx, ly: dimension of cube in x,y direction (Mpc)
    sigma_beam: FWHM of beam (deg)
    cosmo: Astropy cosmology to use in beam smoothing calculations
    '''
    nx,ny,nz = np.shape(dT)
    dpix = np.mean([lx/nx,ly/ny]) # pixel size
    r = cosmo.comoving_distance(zeff)
    R_beam = (np.radians(sigma_beam)/(2*np.sqrt(2*np.log(2))))*r.value*cosmo.h
    for j in range(nz):
        dT[:,:,j] = gaussian_filter(dT[:,:,j], sigma=R_beam/dpix, mode='reflect')
    return dT


def LoSpixels(image_cube, mean_center=True):
    '''
    Convert image data cube from [Nx, Ny, Nz] to [Nz, Npix] format.

    INPUTS:
    image_cube: input data cube in image form, with shape [Nx, Ny, Nz] where Nz
    mean_center: if True, do mean centering for each frequency slice after 
    '''

    # swap axes to [Nz, Nx, Ny] for converting to visibilities:
    image_cube = np.swapaxes(image_cube,0,1)
    image_cube = np.swapaxes(image_cube,0,2)

    # conver to LoS pixels format [Nz, Npix]
    axes = np.shape(image_cube)
    image_LoSpixels = np.reshape(image_cube,(axes[0], axes[1]*axes[2]))
    
    # mean center the data:
    if mean_center==True:
        nz = np.shape(image_LoSpixels)[0]
        for i in range(nz):
            image_LoSpixels[i] = image_LoSpixels[i] - np.mean(image_LoSpixels[i])
    
    return image_LoSpixels



import numpy as np
import pandas as pd
import GPy
from . import obs_tools as obs


'''
This script contains Steve's code for foreground removal using PCA, and our code for GPR FG removal.
'''


def GPRclean(Input, freqs, k_FG, k_21cm, num_restarts=10, NprePCA=0, noise_data=None, zero_noise=False, heteroscedastic=False, invert=False):
    '''
    Runs foreground clean on IM data using GPR.
    ---------
    Arguments
    ---------
    
    INPUT: data cube to be cleaned, in [Nx,Ny,Nz] where Nz is the frequency direction
    
    FREQS: array of frequency range of data
    
    k_FG, k_21cm: choice of foreground and 21cm kernels (can be a sum or product of different kernels)
    
    NUM_RESTARTS: how many times to optimise the GPR regression model.
        [NOTE: If you already know your best fitting kernels (i.e., you're inputting the already
        optimized k_FG and k_21cm kernels), set num_restarts = 0]
        
    NprePCA: 0 if no pre-PCA is desired, otherwise this number is the N_FG number of components used
        in a pre-PCA clean of the data
    
    NOISE_DATA: input here your noise map in [Nx,Ny,Nz] if you have a reasonable estimate from your data, 
        otherwise set to None and use GPR to try to fit your noise
        
    ZERO_NOISE: if True, the noise in your GPR model will be set to zero and fixed. Otherwise it will try to
        fit to noise in your data, in either the heteroscedastic or non-heteroscedastic case. Set to zero if 
        you want to fit your noise with a separate kernel instead, otherwise you will fit to noise twice.
        
    HETEROSCEDASTIC: if True, runs Heteroscedastic regression model (variable noise)
        (Note: you cannot have zero_noise=False and heteroscedastic=True at the same time, set 
        heteroscedastic=False instead in this case).
        
    INVERT: if True, inverts data in the frequency direction
    '''
    axes = np.shape(Input)
    
    # if desired, do a pre-PCA with N_FG=NprePCA removed components
    if NprePCA > 0: Input = PCAclean(Input, N_FG=NprePCA)[0]
    
    # converting [Nx,Ny,Nz] -> [Npix,Nz]
        # note: we mean center the data but not the inidividual noise cube, since we want
        # to extract the real noise variance in our data.
    Input = obs.LoSpixels(np.copy(Input), mean_center=True)
    if noise_data is not None: noise_data = obs.LoSpixels(noise_data, mean_center=False)
    
    # invert frequency axis
    if invert==True: 
        Input = Input[::-1]
        if noise_data is not None: noise_data = noise_data[::-1]
    
    # build your model, input the freq range, the data, and the kernels
    kern = k_FG + k_21cm
    
    # this heteroscedastic case assumes a Gaussian noise variance that changes with frequency
    if heteroscedastic==True: 
        # this case assumes noise is known, sets noise level to your noise_data variances
            # at different frequencies (since heteroscedastic)
        if noise_data is not None:
            model = GPy.models.GPHeteroscedasticRegression(freqs[:, np.newaxis], Input, kern)
            model.het_Gauss.variance.constrain_fixed(noise_data.var(axis=1)[:, None])
        # this case assumes noise is not known, model will fit a variance at each frequency
        else:
            model = GPy.models.GPHeteroscedasticRegression(freqs[:, np.newaxis], Input, kern)
        # note: if you want the case of *no noise*, there's no need to use heteroscedastic,
            # so set heteroscedastic = False and see below
    
    # this non-heteroscedastic case assumes constant Gaussian noise variance throughout frequency
    else: 
        # this case assumes noise is know, sets the noise variance level to the variance
            # from the input noise_data
        if noise_data is not None:
            model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
            model.Gaussian_noise.constrain_fixed(noise_data.var())
        else:
            # this case assumes there is no noise in your data
            if zero_noise == True:
                model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
                model['.*Gaussian_noise'] = 0.0
                model['.*noise'].fix()
            # this case assumes there is noise but it is unknown, fits a constant variance
            else:
                model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
    
    # optimise model, find best fitting hyperparameters for kernels
    model.optimize_restarts(num_restarts=num_restarts)
    
    # extract optimised foreground kernel (depending on how many kernels were considered)
    if k_FG.name == 'sum':
        k_FG_len = len(k_FG.parts)
        k_FG = model.kern.parts[0]
        if k_FG_len > 1:
            for i in range(1, k_FG_len):
                k_FG += model.kern.parts[i]
    else: k_FG = model.kern.parts[0]
    
    # make prediction of what FGs would look like using this optimised FG kernel
    y_fit, y_cov = model.predict(freqs[:, np.newaxis], full_cov=True, kern=k_FG,
        include_likelihood=False)
    
    # subtract FG fit from data, to obtain HI residuals:
    y_sub = Input - y_fit
    
    # un-revert frequency axis:
    if invert==True: 
        y_sub = y_sub[::-1]
        y_fit = y_fit[::-1]
    
    # reshape foreground fit
    y_fit = np.swapaxes(y_fit,0,1)
    y_fit = np.reshape(y_fit,(axes[0], axes[1], axes[2]))
    
    # reshape residuals
    y_sub = np.swapaxes(y_sub,0,1)
    y_sub = np.reshape(y_sub,(axes[0], axes[1], axes[2]))
    
    # create series as output object
    d = {'res': y_sub, 'fgcov': y_cov, 'fgfit': y_fit, 'model': model}
    result = pd.Series(d)
    
    return result


def GPRfit(Input, freqs, kern, num_restarts=10, NprePCA=0, noise_data=None, zero_noise=False, heteroscedastic=False, invert=False):
    '''
    Allows you to fit any kernel to any data, and obtain a prediction in your frequency range
    (e.g. prediction of what your foregrounds look like, but when fitting ONLY to the foreground data)
    ---------
    Arguments
    ---------
    
    INPUT: data cube to be cleaned, in [Nx,Ny,Nz] where Nz is the frequency direction
    
    FREQS: array of frequency range of data
    
    kern: choice of kernel
    
    NUM_RESTARTS: how many times to optimise the GPR regression model.
        [NOTE: If you already know your best fitting kernels (i.e., you're inputting the already
        optimized k_FG and k_21cm kernels), set num_restarts = 0]
        
    NprePCA: 0 if no pre-PCA is desired, otherwise this number is the N_FG number of components used
        in a pre-PCA clean of the data
    
    NOISE_DATA: input here your noise map in [Nx,Ny,Nz] if you have a reasonable estimate from your data, 
        otherwise set to None and use GPR to try to fit your noise
        
    ZERO_NOISE: if True, the noise in your GPR model will be set to zero and fixed. Otherwise it will try to
        fit to noise in your data, in either the heteroscedastic or non-heteroscedastic case. Set to zero if 
        you want to fit your noise with a separate kernel instead, otherwise you will fit to noise twice.
        
    HETEROSCEDASTIC: if True, runs Heteroscedastic regression model (variable noise)
        (Note: you cannot have zero_noise=False and heteroscedastic=True at the same time, set 
        heteroscedastic=False instead in this case).
        
    INVERT: if True, inverts data in the frequency direction
    '''
    axes = np.shape(Input)
    
    # if desired, do a pre-PCA with N_FG=NprePCA removed components
    if NprePCA > 0: Input = PCAclean(Input, N_FG=NprePCA)[0]
    
    # converting [Nx,Ny,Nz] -> [Npix,Nz]
        # note: we mean center the data but not the inidividual noise cube, since we want
        # to extract the real noise variance in our data.
    Input = obs.LoSpixels(np.copy(Input), mean_center=True)
    if noise_data is not None: noise_data = obs.LoSpixels(noise_data, mean_center=False)
    
    # invert frequency axis
    if invert==True: 
        Input = Input[::-1]
        if noise_data is not None: noise_data = noise_data[::-1]
    
    # this heteroscedastic case assumes a Gaussian noise variance that changes with frequency
    if heteroscedastic==True: 
        # this case assumes noise is known, sets noise level to your noise_data variances
            # at different frequencies (since heteroscedastic)
        if noise_data is not None:
            model = GPy.models.GPHeteroscedasticRegression(freqs[:, np.newaxis], Input, kern)
            model.het_Gauss.variance.constrain_fixed(noise_data.var(axis=1)[:, None])
        # this case assumes noise is not known, model will fit a variance at each frequency
        else:
            model = GPy.models.GPHeteroscedasticRegression(freqs[:, np.newaxis], Input, kern)
        # note: if you want the case of *no noise*, there's no need to use heteroscedastic,
            # so set heteroscedastic = False and see below
    
    # this non-heteroscedastic case assumes constant Gaussian noise variance throughout frequency
    else: 
        # this case assumes noise is know, sets the noise variance level to the variance
            # from the input noise_data
        if noise_data is not None:
            model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
            model.Gaussian_noise.constrain_fixed(noise_data.var())
        else:
            # this case assumes there is no noise in your data
            if zero_noise == True:
                model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
                model['.*Gaussian_noise'] = 0.0
                model['.*noise'].fix()
            # this case assumes there is noise but it is unknown, fits a constant variance
            else:
                model = GPy.models.GPRegression(freqs[:, np.newaxis], Input, kern)
    
    # optimise model, find best fitting hyperparameters for kernels
    model.optimize_restarts(num_restarts=num_restarts)
    
    # make prediction of what FGs would look like using this optimised FG kernel
    y_fit, y_cov = model.predict(freqs[:, np.newaxis], full_cov=True, kern=model.kern,
        include_likelihood=False)
    
    # un-revert frequency axis:
    if invert==True: 
        y_fit = y_fit[::-1]
    
    # reshape residuals
    y_fit = np.swapaxes(y_fit,0,1)
    y_fit = np.reshape(y_fit,(axes[0], axes[1], axes[2]))
    
    # create series as output object
    d = {'cov': y_cov, 'fit': y_fit, 'model': model}
    result = pd.Series(d)
    
    return result


def PCAclean(Input, N_FG=4):
    '''
    Takes input in [Nx,Ny,Nz] data cube form where Nz is number of redshift 
    (frequency) bins. N_FG is number of eigenmodes for PCA to remove
    '''
    
    # Collapse data cube to [Npix,Nz] structure:
    axes = np.shape(Input)
    Input = obs.LoSpixels(Input, mean_center=True)
    Nz,Npix = np.shape(Input)
    
    # Obtain frequency covariance matrix for input data
    C = np.cov(Input)
    eigenval, eigenvec = np.linalg.eigh(C)
    eignumb = np.linspace(1,len(eigenval),len(eigenval))
    eigenval = eigenval[::-1] #Put largest eigenvals first
    A = eigenvec[:,Nz-N_FG:] # Mixing matrix
    S = np.dot(A.T,Input)
    
    # PCA Component Maps
    Recon_FG = np.dot(A,S)
    Residual = Input - Recon_FG
    Residual = np.swapaxes(Residual,0,1) #[Nz,Npix]->[Npix,Nz]
    Residual = np.reshape(Residual,(axes[0],axes[1],axes[2]))
    
    return Residual,eignumb,eigenval



'''
Setup script, to make package pip installable
'''


from setuptools import setup


# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(name = 'gpr4im',
      version = '1.0.1',
      description = 'Python package for using GPR as a foreground removal technique in 21cm intensity mapping',
      author = 'Paula S. Soares',
      author_email = 'p.s.soares@qmul.ac.uk',
      packages = ['gpr4im'],
      url = 'https://github.com/paulassoares/gpr4im',
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      install_requires = ['numpy','scipy','matplotlib','astropy','pandas','GPy','getdist','jupyter'],
)
